#ifndef _CONDITIONS_H_
#define _CONDITIONS_H_

#include <iostream>
#include <vector>

class Point;
class GMPint;
struct grevlexC;
template <typename T, typename Cmp> class Polynom;
class PolynomTask;

class Statement {

public:
	virtual ~Statement();
	virtual void print(std::ostream &out) const = 0;
	virtual void convertToPolynoms(PolynomTask &ptask) const = 0;
};

class Condition : public Statement {
};

class Question: public Statement {

public:
	virtual void postProcessSolution(GMPint &numerator, GMPint &denominator) = 0;
};

class TriangleCondition : public Condition {

public:
	TriangleCondition(const Point &first, const Point &second, const Point &third);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myFirst, mySecond, myThird;

};

class DistanceCondition : public Condition {

public:
	DistanceCondition(const Point &first, const Point &second, const std::string &distanceVariable);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myFirst, mySecond;
	const std::string myDistanceVariable;

};

class DistanceToSegmentCondition : public Condition {

public:
	DistanceToSegmentCondition(const Point &point, const Point &first, const Point &second, const std::string &distanceVariable);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myPoint, myFirst, mySecond;
	const std::string myDistanceVariable;

};

class PointDivideSegmentCondition : public Condition {

public:
	PointDivideSegmentCondition(const Point &point, const Point &left, const Point &right, const std::string &coefVariable);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myPoint, myLeft, myRight;
	const std::string myCoeficientVariable;

};

class MiddlePointCondition : public Condition {

public:
	MiddlePointCondition(const Point &first, const Point &second, const Point &middle);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myFirst, mySecond, myMiddle;

};

class ParallelLinesCondition : public Condition {

public:
	ParallelLinesCondition(const Point &A, const Point &B, const Point &C, const Point &D);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myA, myB, myC, myD;

};

class PointInSegmentCondition : public Condition {

public:
	PointInSegmentCondition(const Point &P, const Point &A, const Point &B);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myP, myA, myB;

};

class AngleCondition : public Condition {

public:
	AngleCondition(const Point &left, const Point &vertex, const Point &right, const std::string &angleVariable);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myLeft, myVertex, myRight;
	const std::string myAngleVariable;

};

class TriangleAreaCondition : public Condition {

public:
	TriangleAreaCondition(const Point &first, const Point &second, const Point &third, const std::string &areaVariable);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myFirst, mySecond, myThird;
	const std::string myAreaVariable;

};

class CollinearPointsCondition : public Condition {

public:
	CollinearPointsCondition(const Point &first, const Point &second, const Point &third);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const Point myFirst, mySecond, myThird;

};

class EqualCondition : public Condition {

public:
	EqualCondition(const std::string &leftVar, const std::string &rightVar);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const std::string myLeftVariable, myRightVariable;

};

class TimesCondition : public Condition {

public:
	TimesCondition(const std::string &leftVar, const std::string &rightVar, const std::string &resultVar);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;

private:
	const std::string myLeftVariable, myRightVariable, myResultVariable;

};

class DistanceQuestion : public Question {

public:
	DistanceQuestion(const Point &first, const Point &second);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;
	virtual void postProcessSolution(GMPint &numerator, GMPint &denominator);

private:
	const Point myFirst, mySecond;
};

class AngleQuestion : public Question {

public:
	AngleQuestion(const Point &left, const Point &vertex, const Point &right);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;
	virtual void postProcessSolution(GMPint &numerator, GMPint &denominator);

private:
	const Point myLeft, myVertex, myRight;
};

class TriangleAreaQuestion : public Question {

public:
	TriangleAreaQuestion(const Point &first, const Point &second, const Point &third);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;
	virtual void postProcessSolution(GMPint &numerator, GMPint &denominator);


private:
	const Point myFirst, mySecond, myThird;

};

class TrapeziumAreaQuestion : public Question {

public:
	TrapeziumAreaQuestion(const Point &A, const Point &B, const Point &C, const Point &D);
	virtual void print(std::ostream &out) const;
	virtual void convertToPolynoms(PolynomTask &ptask) const;
	virtual void postProcessSolution(GMPint &numerator, GMPint &denominator);


private:
	const Point myA, myB, myC, myD;

};

#endif // _CONDITIONS_H_
