#ifndef POLYNOM_H_
#define POLYNOM_H_
#include <map>
#include <vector>
#include <set>
#include <list>
#include <stdio.h>
#include <string>
#include <hash_set>
#include <boost/functional/hash.hpp>


struct lexC;
struct grlexC;
struct grevlexC;
class VarDump;

struct pairC;

class Monom {

private:
  std::map <size_t, size_t> myPowers;
  int myPower;

public:
  Monom(const Monom& m);
  Monom();
  Monom& operator=(const Monom& m);

  bool operator==(const Monom& m) const;
  Monom& operator*=(const Monom& m);
  Monom& operator/=(const Monom& m);
  size_t& operator[](size_t n);
  Monom operator*(const Monom& m) const;
  Monom operator/(const Monom& m) const;

  int getPower() const;
  bool divisible(const Monom& m) const;
  bool isOneVariable(size_t num) const;

  friend Monom LCM(const Monom& a, const Monom& b);
  friend class lexC;
  friend class grlexC;
  friend class grevlexC;
  friend class VarDump;

};

struct lexC {
  bool operator()(const Monom& m1, const Monom& m2);
};

struct grlexC {
  bool operator()(const Monom& m1, const Monom& m2);
};

struct grevlexC {
  bool operator()(const Monom& m1, const Monom& m2);
};

template <typename T, typename Cmp>
class Polynombasis;

template <typename T, typename Cmp>
class optBasis;

template <typename T, typename Cmp>
class Polynom {

private:
  std::map <Monom,T, Cmp> myNoms;

public:

  Polynom();

  Polynom(Polynom const& p);
  Polynom(Monom const& m);
  Polynom& operator=(const Polynom& p);
  Polynom& operator=(const Monom& m);

  Polynom& operator*=(const T& i);
  Polynom& operator*=(const Monom& m);
  Polynom& operator+=(const Monom& m);
  Polynom& operator-=(const Monom& m);
  Polynom& operator+=(const T& i);
  Polynom& operator-=(const T& i);
  void addNom(Monom m, const T& k);
  Polynom& operator+=(const Polynom& p);
  Polynom& operator-=(const Polynom& p);
  Polynom& operator*=(const Polynom& p);

  Polynom operator+(const Polynom &p) const;
  Polynom operator-(const Polynom &p) const;
  Polynom operator*(const Polynom &p) const;
  Polynom operator+(const Monom &m) const;
  Polynom operator-(const Monom &m) const;
  Polynom operator*(const Monom &m) const;
  Polynom operator*(const T& i) const;
  Polynom operator-() const;

  bool operator == (const Polynom &p) const;

  Monom LM() const;
  T LC() const;
  bool isNull() const;
  bool isNumber() const;
  Polynom getRemainder(Polynombasis<T, Cmp>& ps) const;
  Polynom getRemainder(Polynombasis<T, Cmp>& ps, T& denominator) const;
  T* getValue(T& value, T& denominator, const Polynombasis<T, Cmp>& ps);
  void reduce();
  bool isOneVariable(size_t num) const;

  friend class VarDump;

};

template <typename T, typename Cmp>
class Polynombasis {

private:
  std::vector<Polynom<T, Cmp> > myPolynoms;
  void reduce();

public:
  Polynombasis();
  Polynombasis(const Polynombasis& p);
  Polynombasis& operator=(const Polynombasis& p);

  const Polynom<T, Cmp>& operator[](size_t i) const;
  void addPolynom(const Polynom<T, Cmp> &p);
  Polynombasis makeBasis() const;
  Polynombasis makeBasis_h() const;
  bool contains(const Polynom<T, Cmp> &p) const;
  size_t getSize() const;
  void sort();
  bool isNumber();

  friend class VarDump;

};

struct Basismonom;

struct Basispair {

public:
  size_t I, J;
  size_t IJ;
  std::multiset<Basismonom>::iterator toMonom;
  bool operator<(const Basispair& p) const;

};

struct Basismonom {

public:
  Monom M;
  std::set<Basispair>::iterator toPair;
  bool operator<(const Basismonom& m) const;

};

class Pairs {

private:
  std::set<Basispair> myPairs;
  std::multiset<Basismonom> myMonoms;
  size_t mySize;

public:
  Pairs():mySize(0){};

  void add(size_t i, size_t j, const Monom& m);
  size_t getfirstI() const;
  size_t getfirstJ() const;
  Monom getfirstM() const;
  void removefirst();
  bool empty() const;
  bool contains(size_t i, size_t j) const;
/*  void removenum(size_t num);*/

};

struct Basismonom_h;

struct Basispair_h {

public:
  std::pair<size_t, size_t> P;
  std::multiset<Basismonom_h>::iterator toMonom;
  bool operator==(const Basispair_h& p) const;

};

struct Pairhash {
  size_t operator()(const Basispair_h& p) const{
    return (boost::hash<std::pair<size_t, size_t> >()(p.P));
  }
};

struct Basismonom_h {

public:
  Monom M;
  __gnu_cxx::hash_set<Basispair_h,Pairhash>::iterator toPair;
  bool operator<(const Basismonom_h& m) const;

};

class Pairs_h {

private:
  __gnu_cxx::hash_set<Basispair_h,Pairhash> myPairs;
  std::multiset<Basismonom_h> myMonoms;
  size_t mySize;

public:
  Pairs_h():mySize(0){};

  void add(size_t i, size_t j, const Monom& m);
  size_t getfirstI() const;
  size_t getfirstJ() const;
  Monom getfirstM() const;
  void removefirst();
  bool empty() const;
  bool contains(size_t i, size_t j) const;

};

class VarDump {

private:
  std::vector<std::string> myData;

public:
  size_t addVar(const std::string& s);  //add variable
  size_t getNum(const std::string& s) const;   //get variable number or -1
  const std::string& getVar(size_t n) const;   //get variable by number

  std::string Monomtostr(const Monom& m) const;
  Monom strtoMonom(const std::string& s);

  template <typename T, typename Cmp>
  std::string Polynomtostr(const Polynom<T, Cmp>& p) const;  //должны быть пробелы перед + и - и не должны быть после

  template <typename T, typename Cmp>
  Polynom<T, Cmp> strtoPolynom(const std::string& s);

  template <typename T, typename Cmp>
  Polynom<T, Cmp> freadPolynom(FILE* f);

  template <typename T, typename Cmp>
  void printPolynom(const Polynom<T, Cmp>& p);

  template <typename T, typename Cmp>
  void fprintPolynom(FILE* f, const Polynom<T, Cmp>& p);

  template <typename T, typename Cmp>
  void fprintBasis(FILE* f, const Polynombasis<T, Cmp>& p);

  template <typename T, typename Cmp>
  void printBasis(const Polynombasis<T, Cmp>& p);

};

Monom LCM(Monom& a, Monom& b);
template <typename T, typename Cmp>
Polynom<T, Cmp> S(const Polynom<T, Cmp> &f, const Polynom<T, Cmp> &g);

template <typename T>
Polynom<T, lexC> getPolynom(const Polynombasis<T, lexC>& ps, const Polynom<T, lexC>& p, size_t num);
#endif
