#ifndef GMPINT_H_
#define GMPINT_H_
#include <gmp.h>
#include <iostream>

class GMPint {

private:
  mpz_t myValue;

public:
  GMPint();
  GMPint(const GMPint &g);
  GMPint(long int i);
  ~GMPint();
  GMPint& operator=(const GMPint &g);
  GMPint& operator=(long int i);

  GMPint& operator+=(const GMPint& g);
  GMPint& operator-=(const GMPint& g);
  GMPint& operator*=(const GMPint& g);
  GMPint& operator/=(const GMPint& g);
  GMPint& operator%=(const GMPint& g);
  GMPint operator+(const GMPint &g) const;
  GMPint operator-(const GMPint &g) const;
  GMPint operator*(const GMPint &g) const;
  GMPint operator/(const GMPint &g) const;
  GMPint operator%(const GMPint &g) const;
  bool operator==(const GMPint &g) const;
  bool operator!=(const GMPint &g) const;
  bool operator<(const GMPint &g) const;
  bool operator>(const GMPint &g) const;
  bool operator<=(const GMPint &g) const;
  bool operator>=(const GMPint &g) const;

  GMPint operator-() const;

  GMPint& operator+=(long int i);
  GMPint& operator-=(long int i);
  GMPint& operator*=(long int i);
  GMPint& operator/=(long int i);
  GMPint& operator%=(long int i);
  GMPint operator+(long int i) const;
  GMPint operator-(long int i) const;
  GMPint operator*(long int i) const;
  GMPint operator/(long int i) const;
  GMPint operator%(long int i) const;
  bool operator==(long int i) const;
  bool operator!=(long int i) const;
  bool operator<(long int i) const;
  bool operator>(long int i) const;
  bool operator<=(long int i) const;
  bool operator>=(long int i) const;

  GMPint& operator++();
  GMPint& operator--();

  long int getlongint() const;

  void fprint(FILE* f) const;
  void fread(FILE* f) const;

  friend GMPint gcd(const GMPint& a, const GMPint& b);
  friend std::ostream& operator<< (std::ostream& o, const GMPint& g);
};

std::ostream& operator<< (std::ostream& o, const GMPint& g);

GMPint gcd(const GMPint& a, const GMPint& b);
const GMPint& max(const GMPint& a, const GMPint& b);

//-----------------------------------------------------------------------------------

inline GMPint::GMPint() {
  mpz_init(myValue);
}

inline GMPint::GMPint(const GMPint &g) {
  mpz_init(myValue);
  mpz_set(myValue, g.myValue);
}

inline GMPint::GMPint(long int i) {
  mpz_init(myValue);
  mpz_set_si(myValue, i);
}

inline GMPint::~GMPint() {
  mpz_clear(myValue);
}

inline GMPint& GMPint::operator=(const GMPint &g) {
  mpz_set(myValue, g.myValue);
  return *this;
}

inline GMPint& GMPint::operator=(long int i) {
  mpz_set_si(myValue, i);
  return *this;
}

inline GMPint& GMPint::operator+=(const GMPint& g) {
  mpz_add(myValue, myValue, g.myValue);
  return *this;
}

inline GMPint GMPint::operator+(const GMPint& g) const {
  GMPint temp;
  mpz_add(temp.myValue, myValue, g.myValue);
  return temp;
}

inline GMPint& GMPint::operator-=(const GMPint& g) {
  mpz_sub(myValue, myValue, g.myValue);
  return *this;
}

inline GMPint GMPint::operator-(const GMPint& g) const {
  GMPint temp;
  mpz_sub(temp.myValue, myValue, g.myValue);
  return temp;
}
inline GMPint& GMPint::operator*=(const GMPint& g) {
  mpz_mul(myValue, myValue, g.myValue);
  return *this;
}

inline GMPint GMPint::operator*(const GMPint& g) const {
  GMPint temp;
  mpz_mul(temp.myValue, myValue, g.myValue);
  return temp;
}
inline GMPint& GMPint::operator/=(const GMPint& g) {
  mpz_tdiv_q(myValue, myValue, g.myValue);
  return *this;
}

inline GMPint GMPint::operator/(const GMPint& g) const {
  GMPint temp;
  mpz_tdiv_q(temp.myValue, myValue, g.myValue);
  return temp;
}
inline GMPint& GMPint::operator%=(const GMPint& g) {
  mpz_tdiv_r(myValue, myValue, g.myValue);
  return *this;
}

inline GMPint GMPint::operator%(const GMPint& g) const {
  GMPint temp;
  mpz_tdiv_r(temp.myValue, myValue, g.myValue);
  return temp;
}

inline bool GMPint::operator==(const GMPint &g) const {
  return (mpz_cmp(myValue, g.myValue)==0);
}

inline bool GMPint::operator!=(const GMPint &g) const {
  return (mpz_cmp(myValue, g.myValue)!=0);
}

inline bool GMPint::operator>(const GMPint &g) const {
  return (mpz_cmp(myValue, g.myValue)>0);
}

inline bool GMPint::operator<(const GMPint &g) const {
  return (mpz_cmp(myValue, g.myValue)<0);
}

inline bool GMPint::operator>=(const GMPint &g) const {
  return (mpz_cmp(myValue, g.myValue)>=0);
}

inline bool GMPint::operator<=(const GMPint &g) const {
  return (mpz_cmp(myValue, g.myValue)<=0);
}

inline GMPint GMPint::operator-() const{
  GMPint temp;
  mpz_neg(temp.myValue, myValue);
  return temp;
}

inline GMPint& GMPint::operator+=(long int i) {
  if (i >= 0 ) {
    mpz_add_ui(myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_sub_ui(myValue, myValue, -i);
  }
  return *this;
}

inline GMPint GMPint::operator+(long int i) const {
  GMPint temp;
  if (i >= 0 ) {
    mpz_add_ui(temp.myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_sub_ui(temp.myValue, myValue, -i);
  }
  return temp;
}

inline GMPint& GMPint::operator-=(long int i) {
  if (i >= 0 ) {
    mpz_sub_ui(myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_add_ui(myValue, myValue, -i);
  }
  return *this;
}

inline GMPint GMPint::operator-(long int i) const {
  GMPint temp;
  if (i >= 0 ) {
    mpz_sub_ui(temp.myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_add_ui(temp.myValue, myValue, -i);
  }
  return temp;
}

inline GMPint& GMPint::operator*=(long int i) {
  mpz_mul_si(myValue, myValue, i);
  return *this;
}

inline GMPint GMPint::operator*(long int i) const {
  GMPint temp;
  mpz_mul_si(temp.myValue, myValue, i);
  return temp;
}

inline GMPint& GMPint::operator/=(long int i) {
  if (i >= 0 ) {
    mpz_tdiv_q_ui(myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_tdiv_q_ui(myValue, myValue, -i);
    mpz_neg(myValue, myValue);
  }
  return *this;
}

inline GMPint GMPint::operator/(long int i) const {
  GMPint temp;
  if (i >= 0 ) {
    mpz_tdiv_q_ui(temp.myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_tdiv_q_ui(temp.myValue, myValue, -i);
    mpz_neg(temp.myValue, temp.myValue);
  }
  return temp;
}

inline GMPint& GMPint::operator%=(long int i) {
  if (i >= 0 ) {
    mpz_tdiv_r_ui(myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_tdiv_r_ui(myValue, myValue, -i);
  }
  return *this;
}

inline GMPint GMPint::operator%(long int i) const {
  GMPint temp;
  if (i >= 0 ) {
    mpz_tdiv_r_ui(temp.myValue, myValue, i);
  }
  if (i < 0 ) {
    mpz_tdiv_r_ui(temp.myValue, myValue, -i);
  }
  return temp;
}

inline bool GMPint::operator==(long int i) const {
  return (mpz_cmp_si(myValue, i)==0);
}

inline bool GMPint::operator!=(long int i) const {
  return (mpz_cmp_si(myValue, i)!=0);
}

inline bool GMPint::operator>(long int i) const {
  return (mpz_cmp_si(myValue, i)>0);
}

inline bool GMPint::operator<(long int i) const {
  return (mpz_cmp_si(myValue, i)<0);
}

inline bool GMPint::operator>=(long int i) const {
  return (mpz_cmp_si(myValue, i)>=0);
}

inline bool GMPint::operator<=(long int i) const {
  return (mpz_cmp_si(myValue, i)<=0);
}

inline GMPint& GMPint::operator++() {
  mpz_add_ui(myValue, myValue, 1);
  return *this;
}

inline GMPint& GMPint::operator--() {
  mpz_sub_ui(myValue, myValue, 1);
  return *this;
}

inline long int GMPint::getlongint() const {
  return mpz_get_si(myValue);
}

inline std::ostream& operator<< (std::ostream& o, const GMPint& g) {
  o << g.myValue;
  return o;
}

inline void GMPint::fprint(FILE* f) const {
  gmp_fprintf(f, "%Zd", myValue);
}

inline void GMPint::fread(FILE* f) const {
  gmp_fscanf(f, "%Zd", &myValue);
}

inline GMPint gcd(const GMPint& a, const GMPint& b) {
  GMPint temp;
  mpz_gcd(temp.myValue, a.myValue, b.myValue);
  return temp;
}

inline const GMPint& max(const GMPint& a, const GMPint& b) {
  if (a>=b) return a;
  return b;
}

#endif
