#include <cstdlib>
#include <cmath>
#include <sstream>

#include "gmpint.h"
#include "point.h"
#include "tokenizer.h"
#include "nodehandlers.h"
#include "task.h"
#include "syntaxtree.h"

NodeHandler::~NodeHandler() {
}

bool MiddlePointNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 2;
}

Token MiddlePointNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &mp = task.getAvailablePointName();
	task.addMiddlePointCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(mp));
	return Token(mp, Token::ttIdentifier, node->getToken().getPosition());
}

bool AngleNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token AngleNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	std::string angle = task.getAvailableVariableName();
	task.addAngleCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3), angle);
	return Token(angle, Token::ttIdentifier, node->getToken().getPosition());
}

bool DistanceNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 2 || node->childrenNumber() == 3;
}

Token DistanceNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	std::string dist = task.getAvailableVariableName();
	if (node->childrenNumber() == 2) {
		task.addDistanceCondition(task.getPoint(p1), task.getPoint(p2), dist);
	} else {
		const std::string &p3 = node->getChild(0)->getToken().getValue();
		task.addDistanceToSegmentCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3), dist);
	}
	return Token(dist, Token::ttIdentifier, node->getToken().getPosition());
}

bool TriangleNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token TriangleNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	task.addTriangleCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3));
	return Token(p1 + p2 + p3, Token::ttIdentifier, node->getToken().getPosition());
}

bool TriangleAreaNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token TriangleAreaNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	std::string area = task.getAvailableVariableName();
	task.addTriangleAreaCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3), area);
	return Token(area, Token::ttIdentifier, node->getToken().getPosition());
}

bool EqualNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 2;
}

Token EqualNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &left = node->getChild(0)->getToken().getValue();
	const std::string &right = node->getChild(1)->getToken().getValue();
	task.addEqualCondition(left, right);
	return Token(right, node->getChild(1)->getToken().getType(), node->getToken().getPosition());
}

bool TimesNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 2;
}

Token TimesNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &left = node->getChild(0)->getToken().getValue();
	const std::string &right = node->getChild(1)->getToken().getValue();
	std::string result = task.getAvailableVariableName();
	task.addTimesCondition(left, right, result);
	return Token(result, Token::ttIdentifier, node->getToken().getPosition());
}

bool FindDistanceNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 2;
}

Token FindDistanceNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	task.addDistanceQuestion(task.getPoint(p1), task.getPoint(p2));
	return Token("?", Token::ttNumber, node->getToken().getPosition());
}

bool FindAngleNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token FindAngleNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	task.addAngleQuestion(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3));
	return Token("?", Token::ttNumber, node->getToken().getPosition());
}

bool FindTriangleAreaNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token FindTriangleAreaNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	task.addTriangleAreaQuestion(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3));
	return Token("?", Token::ttNumber, node->getToken().getPosition());
}

bool CosNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 1;
}

Token CosNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &angle = node->getChild(0)->getToken().getValue();
	double pi = 2 * acos(0);
	std::ostringstream cosAngle;
	cosAngle << (int)cos(atoi(angle.c_str()) * pi / 180);
	return Token(cosAngle.str(), Token::ttNumber, node->getToken().getPosition());
}

bool CollinearPointsNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token CollinearPointsNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	task.addCollinearPointsCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3));
	return Token(p1 + p2 + p3, Token::ttIdentifier, node->getToken().getPosition());
}

bool FindTrapeziumAreaNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 4;
}

Token FindTrapeziumAreaNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	const std::string &p4 = node->getChild(3)->getToken().getValue();
	task.addTrapeziumAreaQuestion(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3), task.getPoint(p4));
	return Token("?", Token::ttNumber, node->getToken().getPosition());
}

bool PointInSegmentNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token PointInSegmentNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	task.addPointInSegmentCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3));
	return node->getChild(0)->getToken();
}

bool ParallelLinesNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 4;
}

Token ParallelLinesNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	const std::string &p4 = node->getChild(3)->getToken().getValue();
	task.addParallelLinesCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3), task.getPoint(p4));
	return Token(p1 + p2 + p3 + p4, Token::ttIdentifier, node->getToken().getPosition());
}

bool PointDivideSegmentNodeHandler::isValid(SyntaxTreeNode *node) {
	return node->childrenNumber() == 3;
}

Token PointDivideSegmentNodeHandler::addToModel(Task &task, SyntaxTreeNode *node) {
	const std::string &p1 = node->getChild(0)->getToken().getValue();
	const std::string &p2 = node->getChild(1)->getToken().getValue();
	const std::string &p3 = node->getChild(2)->getToken().getValue();
	std::string coef = task.getAvailableVariableName();
	task.addPointDivideSegmentCondition(task.getPoint(p1), task.getPoint(p2), task.getPoint(p3), coef);
	return Token(coef,  Token::ttIdentifier, node->getToken().getPosition());
}
