#include <stack>
#include <iostream>

#include "tokenizer.h"
#include "syntaxtree.h"
#include "parser.h"

Parser::Parser(Tokenizer &scanner) : myScanner(scanner) {
}

void Parser::printError(std::pair<int,int> position, const std::string &error) const {
	std::cerr << "Error[Line " << position.first << ", Position " << position.second << "]: " << error << std::endl;
}

void Parser::parse(std::vector<SyntaxTreeNode *> &result) const {
	std::stack<SyntaxTreeNode *> stack;
	ParserState state = psStart;
	bool isTokenId;

	while (myScanner.hasTokens()) {
		Token t = myScanner.nextToken();
		if (t.getType() == Token::ttError) {
			printError(t.getPosition(), t.getValue());
			return;
		}

		switch (state) {
			case psStart:
				if (t.getType() == Token::ttIdentifier) {
					state = psWaitForLBracket;
					stack.push(new SyntaxTreeNode(t));
				} else {
					printError(t.getPosition(), "Condition must starts with identifier");
					return;
				}
				break;
			case psWaitForLBracket:
				if (t.getValue() == "(") {
					state = psArgument;
				} else {
					printError(t.getPosition(), "Expected (");
					return;
				}
				break;
			case psArgument:
				if (t.getType() == Token::ttIdentifier || t.getType() == Token::ttNumber) {
					state = psWaitForArgDelimiter;

					SyntaxTreeNode *node = new SyntaxTreeNode(t);
					SyntaxTreeNode *p = stack.top();
					p->addChild(node);
					stack.push(node);

					isTokenId = t.getType() == Token::ttIdentifier;
				} else {
					printError(t.getPosition(), "Argument of condition must be identifier or number");
					return;
				}
				break;
			case psWaitForArgDelimiter:
				if (t.getValue() == ",") {
					state = psArgument;
					stack.pop();
				} else if (t.getValue() == ")") {
					stack.pop();
					if (stack.size() == 1) {
						state = psFinish;
					}
				} else if (t.getValue() == "(" && isTokenId) {
					state = psArgument;
				} else {
					printError(t.getPosition(), "Expected ( or ) or ,");
					return;
				}
				break;
			case psFinish:
				if (t.getValue() == ";") {
					state = psStart;
					result.push_back(stack.top());
					stack.pop();					
				} else if (t.getValue() == "=") {
					state = psWaitForReturnValue;
					SyntaxTreeNode *node = new SyntaxTreeNode(t);
					SyntaxTreeNode *p = stack.top();
					node->addChild(p);
					stack.pop();
					stack.push(node);
				} else {
					printError(t.getPosition(), "Expected ; or =");
					return;
				}
				break;
			case psWaitForReturnValue:
				if (t.getType() == Token::ttIdentifier || t.getType() == Token::ttNumber) {
					state = psWaitForSemicolon;
					SyntaxTreeNode *p = stack.top();
					SyntaxTreeNode *node = new SyntaxTreeNode(t);
					p->addChild(node);
					result.push_back(p);
					stack.pop();
				} else {
					printError(t.getPosition(), "Expected identifier or number");
					return;
				}
				break;
			case psWaitForSemicolon:
				if (t.getValue() == ";") {
					state = psStart;
				} else {
					printError(t.getPosition(), "Expected ;");
					return;
				}
				break;
		}
	}
}
