#include <iostream>
#include <sstream>

#include "point.h"
#include "conditions.h"
#include "task.h"
#include "gmpint.h"
#include "Polynom.h"
#include "Polynom.impl.h"
#include "polynomtask.h"

Task::Task(): myAvailablePointNamePrefix("TemporaryPoint") {
	myAvailablePointNumber = 0;
}

Task::~Task() {
	for (std::vector<Condition *>::iterator i = myConditions.begin(); i != myConditions.end(); ++i) {
		delete *i;
	}
	for (std::vector<Question *>::iterator i = myQuestions.begin(); i != myQuestions.end(); ++i) {
		delete *i;
	}
}

std::string Task::getAvailablePointName() {
	std::map<std::string, Point>::const_iterator it;
	while ((it = myPoints.find(getCurrentAvailablePointName())) != myPoints.end()) {
		++myAvailablePointNumber;
	}
	return getCurrentAvailablePointName();
}

std::string Task::getCurrentAvailablePointName() {
	std::ostringstream oss;
	oss << myAvailablePointNamePrefix << myAvailablePointNumber;
	return oss.str();
}

std::string Task::getAvailableVariableName() {
	return myPolynomTask.getAvailableVariableName();
}

const Point &Task::getPoint(const std::string &name) {
	std::map<std::string, Point>::const_iterator it = myPoints.find(name);
	if (it != myPoints.end()) {
		return it->second;
	} else {
		return myPoints.insert(std::make_pair(name, Point(name))).first->second;
	}
}

void Task::addTriangleCondition(const Point &A, const Point &B, const Point &C) {
	myConditions.push_back(new TriangleCondition(A, B, C));
}

void Task::addDistanceCondition(const Point &A, const Point &B, const std::string &distVar) {
	myConditions.push_back(new DistanceCondition(A, B, distVar));
}

void Task::addDistanceToSegmentCondition(const Point &P, const Point &A, const Point &B, const std::string &distVar) {
	myConditions.push_back(new DistanceToSegmentCondition(P, A, B, distVar));
}

void Task::addMiddlePointCondition(const Point &A, const Point &B, const Point &M) {
	myConditions.push_back(new MiddlePointCondition(A, B, M));
}

void Task::addAngleCondition(const Point &left, const Point &vertex, const Point &right, const std::string &angleVar) {
	myConditions.push_back(new AngleCondition(left, vertex, right, angleVar));
}

void Task::addTriangleAreaCondition(const Point &first, const Point &second, const Point &third, const std::string &areaVar) {
	myConditions.push_back(new TriangleAreaCondition(first, second, third, areaVar));
}

void Task::addPointDivideSegmentCondition(const Point &point, const Point &left, const Point &right, const std::string &coefVariable) {
	myConditions.push_back(new PointDivideSegmentCondition(point, left, right, coefVariable));
}

void Task::addParallelLinesCondition(const Point &A, const Point &B, const Point &C, const Point &D) {
	myConditions.push_back(new ParallelLinesCondition(A, B, C, D));
}

void Task::addEqualCondition(const std::string &leftVar, const std::string &rightVar) {
	myConditions.push_back(new EqualCondition(leftVar, rightVar));
}

void Task::addTimesCondition(const std::string &leftVar, const std::string &rightVar, const std::string &resultVar) {
	myConditions.push_back(new TimesCondition(leftVar, rightVar, resultVar));
}

void Task::addPointInSegmentCondition(const Point &P, const Point &A, const Point &B) {
	myConditions.push_back(new PointInSegmentCondition(P, A, B));
}

void Task::addCollinearPointsCondition(const Point &first, const Point &second, const Point &third) {
	myConditions.push_back(new CollinearPointsCondition(first, second, third));
}

void Task::addDistanceQuestion(const Point &A, const Point &B) {
	myQuestions.push_back(new DistanceQuestion(A, B));
}

void Task::addAngleQuestion(const Point &left, const Point &vertex, const Point &right) {
	myQuestions.push_back(new AngleQuestion(left, vertex, right));
}

void Task::addTriangleAreaQuestion(const Point &first, const Point &second, const Point &third) {
	myQuestions.push_back(new TriangleAreaQuestion(first, second, third));
}

void Task::addTrapeziumAreaQuestion(const Point &A, const Point &B, const Point &C, const Point &D) {
	myQuestions.push_back(new TrapeziumAreaQuestion(A, B, C, D));
}

std::ostream &operator<<(std::ostream &out, const Task &task) {
	out << "Points: ";
	for (std::map<std::string, Point>::const_iterator i = task.myPoints.begin(); i != task.myPoints.end(); ++i) {
		out << i->second.getName() << " ";
	}
	out << std::endl;

	out << "Conditions:" << std::endl;
	for (std::vector<Condition *>::const_iterator i = task.myConditions.begin(); i != task.myConditions.end(); ++i) {
		(*i)->print(out);
	}

	out << "Questions:" << std::endl;
	for (std::vector<Question *>::const_iterator i = task.myQuestions.begin(); i != task.myQuestions.end(); ++i) {
		(*i)->print(out);
	}

	return out;
}

std::vector<std::pair<GMPint, GMPint> > Task::solve() {
	std::vector<std::pair<GMPint, GMPint> > answers;
	for (std::vector<Condition *>::const_iterator i = myConditions.begin(); i != myConditions.end(); ++i) {
		(*i)->convertToPolynoms(myPolynomTask);
	}
	for (std::vector<Question *>::const_iterator i = myQuestions.begin(); i != myQuestions.end(); ++i) {
		(*i)->convertToPolynoms(myPolynomTask);
		std::pair<GMPint, GMPint> value = myPolynomTask.getValue();
		if (value.second != 0) {
			(*i)->postProcessSolution(value.first,value.second );
			GMPint g = gcd(value.first, value.second);
			value.first /= g;
			value.second /=g;
		}
		answers.push_back(value);
	}
	return answers;
}
