#include <cstring>

#include "tokenizer.h"

Token::Token(const std::string &value, TokenType type, std::pair<size_t, size_t> position) : myValue(value), myType(type), myPosition(position) {
}

const std::string &Token::getValue() const {
	return myValue;
}

Token::TokenType Token::getType() const {
	return myType;
}

std::pair<int, int> Token::getPosition() const {
	return myPosition;
}

void Token::print() const {
	std::string type;
	switch (myType) {
		case ttIdentifier:
			type = "Identifier";
			break;
		case ttNumber:
			type = "Number";
			break;
		case ttDelimiter:
			type = "Delimiter";
			break;
		default:
			type = "Error";
	}
	std::cout << type << '\t' << myValue << std::endl;
}

Tokenizer::Tokenizer(std::istream &inputStream) : myInputStream(inputStream) {
	scan();
}

bool Tokenizer::isDelimiter(char c) const {
	return strchr("(),=;", c);
}

void Tokenizer::scan() {
	TokenizerState state = tsStart;
	std::string token;
	char c;
	size_t symbolNumber = 0, lineNumber = 1;

	while (myInputStream.get(c)) {
		symbolNumber++;
		if (c == '\n') {
			lineNumber++;
		}

		switch (state) {
			case tsStart:
				token = "";
				if (c == '/') {
					state = tsBeginComment;
				} else if (isDelimiter(c)) {
					token += c;
					myTokens.push_back(Token(token, Token::ttDelimiter, std::make_pair(lineNumber, symbolNumber)));
				} else if (c == '+' || c == '-') {
					state = tsSign;
					token += c;
				} else if (isdigit(c)) {
					state = tsNumber;
					token += c;	
				} else if (isalpha(c)) {
					state = tsIdentifier;
					token += c;
				} else if (!isspace(c)) {
					myTokens.push_back(Token("Expected letter or digit or delimiter or sign", Token::ttError,
					std::make_pair(lineNumber, symbolNumber)));
					return;
				}
			break;
			case tsIdentifier:
				if (isalpha(c) || isdigit(c)) {
					token += c;
				} else if (isspace(c)) {
					state = tsStart;
					myTokens.push_back(Token(token, Token::ttIdentifier, std::make_pair(lineNumber, symbolNumber)));
				} else if (isDelimiter(c)) {
					state = tsStart;
					myTokens.push_back(Token(token, Token::ttIdentifier, std::make_pair(lineNumber, symbolNumber)));
					token = "";
					token += c;
					myTokens.push_back(Token(token, Token::ttDelimiter, std::make_pair(lineNumber, symbolNumber)));
				} else if (c == '/') {
					state = tsBeginComment;
					myTokens.push_back(Token(token, Token::ttIdentifier, std::make_pair(lineNumber, symbolNumber)));
				} else {
					myTokens.push_back(Token("Expected letter or delimiter", Token::ttError,
					std::make_pair(lineNumber, symbolNumber)));
					return;
				}
			break;
			case tsNumber:
				if (isdigit(c)) {
					token += c;
				} else if (isspace(c)) {
					state = tsStart;
					myTokens.push_back(Token(token, Token::ttNumber, std::make_pair(lineNumber, symbolNumber)));
				} else if (isDelimiter(c)) {
					state = tsStart;
					myTokens.push_back(Token(token, Token::ttNumber, std::make_pair(lineNumber, symbolNumber)));
					token = "";
					token += c;
					myTokens.push_back(Token(token, Token::ttDelimiter, std::make_pair(lineNumber, symbolNumber)));
				} else if (c == '/') {
					state = tsBeginComment;
					myTokens.push_back(Token(token, Token::ttNumber, std::make_pair(lineNumber, symbolNumber)));
				} else {
					myTokens.push_back(Token("Expected digit or delimiter", Token::ttError, std::make_pair(lineNumber, symbolNumber)));
					return;
				}
			break;
			case tsSign:
				if (isdigit(c)) {
					state = tsNumber;
					token += c;
				} else {
					myTokens.push_back(Token("Expected digit", Token::ttError, std::make_pair(lineNumber, symbolNumber)));
					return;
				}
			break;
			case tsBeginComment:
				if (c == '*') {
					state = tsWaitForStar;
				} else {
					myTokens.push_back(Token("Expected *", Token::ttError, std::make_pair(lineNumber, symbolNumber)));
				}
			break;
			case tsWaitForStar:
				if (c == '*') {
					state = tsEndComment;
				}
			break;
			case tsEndComment:
				if (c == '/') {
					state = tsStart;
				} else {
					state = tsWaitForStar;
				}
			break;
		}
	}
	if (state == tsBeginComment || state == tsWaitForStar || state == tsEndComment || state == tsSign) {
		myTokens.push_back(Token("Unexpected end of file", Token::ttError, std::make_pair(lineNumber, symbolNumber)));
	}
}

Token Tokenizer::nextToken() {
	Token t = myTokens.front();
	myTokens.pop_front();
	return t;
}

bool Tokenizer::hasTokens() const {
	return myTokens.empty() == false;
}
